"""
Import useful functions from _finite_difference module
"""
from pynumdiff.finite_difference._finite_difference import first_order
from pynumdiff.finite_difference._finite_difference import second_order
